#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "CallMany",
]

import sys
from typing import Any, Callable, List, Literal, Optional, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType


try:
    from testtools import MultipleExceptions
except ImportError:
    # Define MultipleExceptions locally if testtools is not available
    class MultipleExceptions(Exception):  # type: ignore[no-redef]
        """Report multiple exc_info tuples in self.args."""


class CallMany(object):
    """A stack of functions which will all be called on __call__.

    CallMany also acts as a context manager for convenience.

    Functions are called in last pushed first executed order.

    This is used by Fixture to manage its addCleanup feature.
    """

    def __init__(self) -> None:
        self._cleanups: List[
            Tuple[Callable[..., Any], Tuple[Any, ...], dict[str, Any]]
        ] = []

    def push(self, cleanup: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        """Add a function to be called from __call__.

        On __call__ all functions are called - see __call__ for details on how
        multiple exceptions are handled.

        :param cleanup: A callable to call during cleanUp.
        :param *args: Positional args for cleanup.
        :param kwargs: Keyword args for cleanup.
        :return: None
        """
        self._cleanups.append((cleanup, args, kwargs))

    def __call__(
        self, raise_errors: bool = True
    ) -> Optional[
        List[
            Tuple[
                Optional[type[BaseException]],
                Optional[BaseException],
                Optional["TracebackType"],
            ]
        ]
    ]:
        """Run all the registered functions.

        :param raise_errors: Deprecated parameter from before testtools gained
            MultipleExceptions. raise_errors defaults to True. When True
            if exception(s) are raised while running functions, they are
            re-raised after all the functions have run.  If multiple exceptions
            are raised, they are all wrapped into a MultipleExceptions object,
            and that is raised.

            Thus, to catch a specific exception from a function run by __call__,
            you need to catch both the exception and MultipleExceptions, and
            then check within a MultipleExceptions instance for an occurrence of
            the type you wish to catch.
        :return: Either None or a list of the exc_info() for each exception
            that occurred if raise_errors was False.
        """
        cleanups = reversed(self._cleanups)
        self._cleanups = []
        result: List[
            Tuple[
                Optional[type[BaseException]],
                Optional[BaseException],
                Optional["TracebackType"],
            ]
        ] = []
        for cleanup, args, kwargs in cleanups:
            try:
                cleanup(*args, **kwargs)
            except Exception:
                result.append(sys.exc_info())
        if result and raise_errors:
            if 1 == len(result):
                error = result[0]
                exc = error[1]
                if exc is not None:
                    raise exc.with_traceback(error[2])
            else:
                raise MultipleExceptions(*result)
        if not raise_errors:
            return result
        return None

    def __enter__(self) -> "CallMany":
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Literal[False]:
        self()
        return False  # Propagate exceptions from the with body.



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "FakePopen",
    "PopenFixture",
]

import random
import subprocess
import sys
from typing import Any, Callable, Dict, IO, List, Optional, Tuple, Union, Final

from fixtures import Fixture


class _Unpassed:
    """Sentinel type for unpassed arguments."""

    pass


_unpassed: Final = _Unpassed()


class FakeProcess(object):
    """A test double process, roughly meeting subprocess.Popen's contract."""

    def __init__(self, args: Dict[str, Any], info: Dict[str, Any]) -> None:
        self._args = args
        self.stdin: Any = info.get("stdin")
        self.stdout: Any = info.get("stdout")
        self.stderr: Any = info.get("stderr")
        self.pid: int = random.randint(0, 65536)
        self._returncode: int = info.get("returncode", 0)
        self.returncode: Optional[int] = None

    @property
    def args(self) -> Any:
        return self._args["args"]

    def poll(self) -> Optional[int]:
        """Get the current value of FakeProcess.returncode.

        The returncode is None before communicate() and/or wait() are called,
        and it's set to the value provided by the 'info' dictionary otherwise
        (or 0 in case 'info' doesn't specify a value).
        """
        return self.returncode

    def communicate(
        self, input: Optional[Union[bytes, str]] = None, timeout: Optional[float] = None
    ) -> Tuple[Any, Any]:
        self.returncode = self._returncode
        if self.stdin and input:
            self.stdin.write(input)
        if self.stdout:
            out = self.stdout.getvalue()
        else:
            out = ""
        if self.stderr:
            err = self.stderr.getvalue()
        else:
            err = ""
        return out, err

    def __enter__(self) -> "FakeProcess":
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.wait()

    def kill(self) -> None:
        pass

    def wait(
        self, timeout: Optional[float] = None, endtime: Optional[float] = None
    ) -> Optional[int]:
        if self.returncode is None:
            self.communicate()
        return self.returncode


class FakePopen(Fixture):
    """Replace subprocess.Popen.

    Primarily useful for testing, this fixture replaces subprocess.Popen with a
    test double.

    :ivar procs: A list of the processes created by the fixture.
    """

    def __init__(
        self, get_info: Callable[[Dict[str, Any]], Dict[str, Any]] = lambda _: {}
    ) -> None:
        """Create a PopenFixture

        :param get_info: Optional callback to control the behaviour of the
            created process. This callback takes a kwargs dict for the Popen
            call, and should return a dict with any desired attributes.
            Only parameters that are supplied to the Popen call are in the
            dict, making it possible to detect the difference between 'passed
            with a default value' and 'not passed at all'.

            e.g.
            def get_info(proc_args):
                self.assertEqual(subprocess.PIPE, proc_args['stdin'])
                return {'stdin': StringIO('foobar')}

            The default behaviour if no get_info is supplied is for the return
            process to have returncode of None, empty streams and a random pid.

            After communicate() or wait() are called on the process object,
            the returncode is set to whatever get_info returns (or 0 if
            get_info is not supplied or doesn't return a dict with an explicit
            'returncode' key).
        """
        super(FakePopen, self).__init__()
        self.get_info = get_info

    def _setUp(self) -> None:
        self.addCleanup(setattr, subprocess, "Popen", subprocess.Popen)
        subprocess.Popen = self  # type: ignore[assignment,misc]
        self.procs: List[FakeProcess] = []

    # The method has the correct signature so we error appropriately if called
    # wrongly.
    def __call__(
        self,
        args: Union[str, List[str]],
        bufsize: Union[int, _Unpassed] = _unpassed,
        executable: Union[str, None, _Unpassed] = _unpassed,
        stdin: Union[None, int, IO[Any], _Unpassed] = _unpassed,
        stdout: Union[None, int, IO[Any], _Unpassed] = _unpassed,
        stderr: Union[None, int, IO[Any], _Unpassed] = _unpassed,
        preexec_fn: Union[Callable[[], None], None, _Unpassed] = _unpassed,
        close_fds: Union[bool, _Unpassed] = _unpassed,
        shell: Union[bool, _Unpassed] = _unpassed,
        cwd: Union[str, None, _Unpassed] = _unpassed,
        env: Union[Dict[str, str], None, _Unpassed] = _unpassed,
        universal_newlines: Union[bool, _Unpassed] = _unpassed,
        startupinfo: Union[Any, _Unpassed] = _unpassed,
        creationflags: Union[int, _Unpassed] = _unpassed,
        restore_signals: Union[bool, _Unpassed] = _unpassed,
        start_new_session: Union[bool, _Unpassed] = _unpassed,
        pass_fds: Union[Any, _Unpassed] = _unpassed,
        *,
        group: Union[str, int, None, _Unpassed] = _unpassed,
        extra_groups: Union[List[Union[str, int]], None, _Unpassed] = _unpassed,
        user: Union[str, int, None, _Unpassed] = _unpassed,
        umask: Union[int, None, _Unpassed] = _unpassed,
        encoding: Union[str, None, _Unpassed] = _unpassed,
        errors: Union[str, None, _Unpassed] = _unpassed,
        text: Union[bool, None, _Unpassed] = _unpassed,
        pipesize: Union[int, _Unpassed] = _unpassed,
        process_group: Union[int, None, _Unpassed] = _unpassed,
    ) -> FakeProcess:
        if sys.version_info < (3, 11) and not isinstance(process_group, _Unpassed):
            raise TypeError(
                "FakePopen.__call__() got an unexpected keyword argument "
                "'process_group'"
            )

        proc_args = dict(args=args)
        local = locals()
        for param in [
            "bufsize",
            "executable",
            "stdin",
            "stdout",
            "stderr",
            "preexec_fn",
            "close_fds",
            "shell",
            "cwd",
            "env",
            "universal_newlines",
            "startupinfo",
            "creationflags",
            "restore_signals",
            "start_new_session",
            "pass_fds",
            "group",
            "extra_groups",
            "user",
            "umask",
            "encoding",
            "errors",
            "text",
            "pipesize",
            "process_group",
        ]:
            if not isinstance(local[param], _Unpassed):
                proc_args[param] = local[param]
        proc_info = self.get_info(proc_args)
        result = FakeProcess(proc_args, proc_info)
        self.procs.append(result)
        return result


PopenFixture = FakePopen



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.


"""Included fixtures."""

__all__ = [
    "ByteStream",
    "DetailStream",
    "EnvironmentVariable",
    "EnvironmentVariableFixture",
    "FakeLogger",
    "FakePopen",
    "LoggerFixture",
    "LogHandler",
    "MockPatch",
    "MockPatchMultiple",
    "MockPatchObject",
    "MonkeyPatch",
    "NestedTempfile",
    "PackagePathEntry",
    "PopenFixture",
    "PythonPackage",
    "PythonPathEntry",
    "StringStream",
    "TempDir",
    "TempHomeDir",
    "Timeout",
    "TimeoutException",
    "WarningsCapture",
    "WarningsFilter",
]


from fixtures._fixtures.environ import (
    EnvironmentVariable,
    EnvironmentVariableFixture,
)
from fixtures._fixtures.logger import (
    FakeLogger,
    LoggerFixture,
    LogHandler,
)
from fixtures._fixtures.mockpatch import (
    MockPatch,
    MockPatchMultiple,
    MockPatchObject,
)
from fixtures._fixtures.monkeypatch import MonkeyPatch
from fixtures._fixtures.popen import (
    FakePopen,
    PopenFixture,
)
from fixtures._fixtures.packagepath import PackagePathEntry
from fixtures._fixtures.pythonpackage import PythonPackage
from fixtures._fixtures.pythonpath import PythonPathEntry
from fixtures._fixtures.streams import (
    ByteStream,
    DetailStream,
    StringStream,
)
from fixtures._fixtures.tempdir import (
    NestedTempfile,
    TempDir,
)
from fixtures._fixtures.temphomedir import (
    TempHomeDir,
)
from fixtures._fixtures.timeout import (
    Timeout,
    TimeoutException,
)
from fixtures._fixtures.warnings import (
    WarningsCapture,
    WarningsFilter,
)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "MonkeyPatch",
]

import functools
import types
from typing import Any

from fixtures import Fixture


_class_types = (type,)


def _coerce_values(
    obj: Any, name: str, new_value: Any, sentinel: Any
) -> tuple[Any, Any]:
    """Return an adapted (new_value, old_value) for patching obj.name.

    setattr transforms a function into an instancemethod when set on a class.
    This checks if the attribute to be replaced is a callable descriptor -
    staticmethod, classmethod, or types.FunctionType - and wraps new_value if
    necessary.

    This also checks getattr(obj, name) and wraps it if necessary
    since the staticmethod wrapper isn't preserved.

    :param obj: The object with an attribute being patched.
    :param name: The name of the attribute being patched.
    :param new_value: The new value to be assigned.
    :param sentinel: If no old_value existed, the sentinel is returned to
        indicate that.
    """
    old_value = getattr(obj, name, sentinel)
    if not isinstance(obj, _class_types):
        # We may be dealing with an instance of a class. In that case the
        # attribute may be the result of a descriptor lookup (or a __getattr__
        # override etc). Its incomplete, but generally good enough to just
        # look and see if name is in the instance dict.
        try:
            obj.__dict__[name]
        except (AttributeError, KeyError):
            return (new_value, sentinel)
        else:
            return (new_value, old_value)

    # getattr() returns a function, this access pattern will return a
    # staticmethod/classmethod if the name method is defined that way
    old_attribute = obj.__dict__.get(name)
    if old_attribute is not None:
        old_value = old_attribute

    # If new_value is not callable, no special handling is needed.
    # (well, technically the same descriptor issue can happen with
    # user supplied descriptors, but that is arguably a feature - someone can
    # deliberately install a different descriptor.
    if not callable(new_value):
        return (new_value, old_value)

    if isinstance(old_value, staticmethod):
        new_value = staticmethod(new_value)
    elif isinstance(old_value, classmethod):
        new_value = classmethod(new_value)
    elif isinstance(old_value, types.FunctionType):
        if hasattr(new_value, "__get__"):
            # new_value is a descriptor, and that would result in it being
            # rebound if we assign it to a class - we want to preserve the
            # bound state rather than having it bound to the new object
            # it has been patched onto.
            captured_method = new_value

            @functools.wraps(old_value)
            def avoid_get(*args: Any, **kwargs: Any) -> Any:
                return captured_method(*args, **kwargs)

            new_value = avoid_get

    return (new_value, old_value)


class MonkeyPatch(Fixture):
    """Replace or delete an attribute."""

    delete = object()

    def __init__(self, name: str, new_value: Any = None) -> None:
        """Create a MonkeyPatch.

        :param name: The fully qualified object name to override.
        :param new_value: A value to set the name to. If set to
            MonkeyPatch.delete the attribute will be deleted.

        During setup the name will be deleted or assigned the requested value,
        and this will be restored in cleanUp.

        When patching methods, the call signature of name should be a subset
        of the parameters which can be used to call new_value.

        For instance.

        >>> class T:
        ...     def method(self, arg1):
        ...         pass
        >>> class N:
        ...     @staticmethod
        ...     def newmethod(arg1):
        ...         pass

        Patching N.newmethod on top of T.method and then calling T().method(1)
        will not work because they do not have compatible call signatures -
        self will be passed to newmethod because the callable (N.newmethod)
        is placed onto T as a regular function. This allows capturing all the
        supplied parameters while still consulting local state in your
        new_value.
        """
        Fixture.__init__(self)
        self.name = name
        self.new_value = new_value

    def _setUp(self) -> None:
        location, attribute = self.name.rsplit(".", 1)
        # Import, swallowing all errors as any element of location may be
        # a class or some such thing.
        try:
            __import__(location, {}, {})
        except ImportError:
            pass

        components = location.split(".")
        current = __import__(components[0], {}, {})
        for component in components[1:]:
            current = getattr(current, component)
        sentinel = object()
        new_value, old_value = _coerce_values(
            current, attribute, self.new_value, sentinel
        )

        if self.new_value is self.delete:
            if old_value is not sentinel:
                delattr(current, attribute)
        else:
            setattr(current, attribute, new_value)

        if old_value is sentinel:
            self.addCleanup(self._safe_delete, current, attribute)
        else:
            self.addCleanup(setattr, current, attribute, old_value)

    def _safe_delete(self, obj: Any, attribute: str) -> Any:
        """Delete obj.attribute handling the case where its missing."""
        sentinel = object()
        if getattr(obj, attribute, sentinel) is not sentinel:
            delattr(obj, attribute)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

from __future__ import annotations

from logging import StreamHandler, getLogger, INFO, Formatter, Handler, LogRecord
import sys
from typing import IO, Optional, Type, TYPE_CHECKING

from fixtures import Fixture
from fixtures._fixtures.streams import StringStream

# Type alias for proper typing during type checking
if TYPE_CHECKING:
    StreamHandlerStr = StreamHandler[IO[str]]
else:
    StreamHandlerStr = StreamHandler

__all__ = [
    "FakeLogger",
    "LoggerFixture",
    "LogHandler",
]


class LogHandler(Fixture):
    """Replace a logger's handlers."""

    def __init__(
        self,
        handler: Handler,
        name: str = "",
        level: Optional[int] = None,
        nuke_handlers: bool = True,
    ) -> None:
        """Create a LogHandler fixture.

        :param handler: The handler to replace other handlers with.
            If nuke_handlers is False, then added as an extra handler.
        :param name: The name of the logger to replace. Defaults to "".
        :param level: The log level to set, defaults to not changing the level.
        :param nuke_handlers: If True remove all existing handles (prevents
            existing messages going to e.g. stdout). Defaults to True.
        """
        super(LogHandler, self).__init__()
        self.handler = handler
        self._name = name
        self._level = level
        self._nuke_handlers = nuke_handlers

    def _setUp(self) -> None:
        logger = getLogger(self._name)
        if self._level:
            self.addCleanup(logger.setLevel, logger.level)
            logger.setLevel(self._level)
        if self._nuke_handlers:
            for handler in reversed(logger.handlers):
                self.addCleanup(logger.addHandler, handler)
                logger.removeHandler(handler)
        try:
            logger.addHandler(self.handler)
        finally:
            self.addCleanup(logger.removeHandler, self.handler)


class StreamHandlerRaiseException(StreamHandlerStr):
    """Handler class that will raise an exception on formatting errors."""

    def handleError(self, record: LogRecord) -> None:
        _, value, tb = sys.exc_info()
        if value is not None:
            raise value.with_traceback(tb)


class FakeLogger(Fixture):
    """Replace a logger and capture its output."""

    def __init__(
        self,
        name: str = "",
        level: int = INFO,
        format: Optional[str] = None,
        datefmt: Optional[str] = None,
        nuke_handlers: bool = True,
        formatter: Optional[Type[Formatter]] = None,
    ) -> None:
        """Create a FakeLogger fixture.

        :param name: The name of the logger to replace. Defaults to "".
        :param level: The log level to set, defaults to INFO.
        :param format: Logging format to use. Defaults to capturing supplied
            messages verbatim.
        :param datefmt: Logging date format to use.
            Mirrors the datefmt used in python loggging.
        :param nuke_handlers: If True remove all existing handles (prevents
            existing messages going to e.g. stdout). Defaults to True.
        :param formatter: a custom log formatter class. Use this if you want
            to use a log Formatter other than the default one in python.

        Example:

          def test_log(self)
              fixture = self.useFixture(LoggerFixture())
              logging.info('message')
              self.assertEqual('message', fixture.output)
        """
        super(FakeLogger, self).__init__()
        self._name = name
        self._level = level
        self._format = format
        self._datefmt = datefmt
        self._nuke_handlers = nuke_handlers
        self._formatter = formatter

    def _setUp(self) -> None:
        name = "pythonlogging:'%s'" % self._name
        stream_fixture = self.useFixture(StringStream(name))
        output = stream_fixture.stream
        self._output: IO[str] = output
        handler = StreamHandlerRaiseException(output)
        if self._format:
            formatter = self._formatter or Formatter
            handler.setFormatter(formatter(self._format, self._datefmt))
        self.useFixture(
            LogHandler(
                handler,
                name=self._name,
                level=self._level,
                nuke_handlers=self._nuke_handlers,
            )
        )

    @property
    def output(self) -> str:
        self._output.seek(0)
        return self._output.read()

    def reset_output(self) -> None:
        self._output.truncate(0)


LoggerFixture = FakeLogger
